@@ -1,10 +1,12 @@
 //! Data structures for representing decoded wasm modules.
 
-use crate::{EntityRef, ModuleTranslation, PrimaryMap, Tunables, WASM_PAGE_SIZE};
+use crate::{ModuleTranslation, PrimaryMap, Tunables, WASM_PAGE_SIZE};
+use cranelift_entity::{packed_option::ReservedValue, EntityRef};
 use indexmap::IndexMap;
 use serde::{Deserialize, Serialize};
 use std::collections::BTreeMap;
 use std::convert::TryFrom;
+use std::mem;
 use std::ops::Range;
 use wasmtime_types::*;
 
@@ -113,6 +115,18 @@ pub struct MemoryInitializer {
     pub data: Range<u32>,
 }
 
+/// Similar to the above `MemoryInitializer` but only used when memory
+/// initializers are statically known to be valid.
+#[derive(Clone, Debug, Serialize, Deserialize)]
+pub struct StaticMemoryInitializer {
+    /// The 64-bit offset, in bytes, of where this initializer starts.
+    pub offset: u64,
+
+    /// The range of data to write at `offset`, where these indices are indexes
+    /// into the compiled wasm module's data section.
+    pub data: Range<u32>,
+}
+
 /// The type of WebAssembly linear memory initialization to use for a module.
 #[derive(Clone, Debug, Serialize, Deserialize)]
 pub enum MemoryInitialization {
@@ -158,7 +172,42 @@ pub enum MemoryInitialization {
         /// indices, like those in `MemoryInitializer`, point within a data
         /// segment that will come as an auxiliary descriptor with other data
         /// such as the compiled code for the wasm module.
-        map: PrimaryMap<MemoryIndex, Vec<(u64, Range<u32>)>>,
+        map: PrimaryMap<MemoryIndex, Vec<StaticMemoryInitializer>>,
+    },
+
+    /// Memory initialization is statically known and involves a single `memcpy`
+    /// or otherwise simply making the defined data visible.
+    ///
+    /// To be statically initialized the same requirements as `Paged` must be
+    /// met, namely that everything references a dfeined memory and all data
+    /// segments have a staitcally known in-bounds base (no globals).
+    ///
+    /// This form of memory initialization is a more optimized version of
+    /// `Segmented` where memory can be initialized with one of a few methods:
+    ///
+    /// * First it could be initialized with a single `memcpy` of data from the
+    ///   module to the linear memory.
+    /// * Otherwise techniques like `mmap` are also possible to make this data,
+    ///   which might reside in a compiled module on disk, available immediately
+    ///   in a linear memory's address space.
+    ///
+    /// To facilitate the latter fo these techniques the `try_static_init`
+    /// function below, which creates this variant, takes a host page size
+    /// argument which can page-align everything to make mmap-ing possible.
+    Static {
+        /// The initialization contents for each linear memory.
+        ///
+        /// This array has, for each module's own linear memory, the contents
+        /// necessary to initialize it. If the memory has a `None` value then no
+        /// initialization is necessary (it's zero-filled). Otherwise with
+        /// `Some` the first element of the tuple is the offset in memory to
+        /// start the initialization and the `Range` is the range within the
+        /// final data section of the compiled module of bytes to copy into the
+        /// memory.
+        ///
+        /// The offset, range base, and range end are all guaranteed to be page
+        /// aligned to the page size passed in to `try_static_init`.
+        map: PrimaryMap<MemoryIndex, Option<StaticMemoryInitializer>>,
     },
 }
 
@@ -191,9 +240,9 @@ impl ModuleTranslation<'_> {
         let mut data = self.data.iter();
         let ok = self.module.memory_initialization.init_memory(
             InitMemory::CompileTime(&self.module),
-            &mut |memory, offset, data_range| {
+            &mut |memory, init| {
                 let data = data.next().unwrap();
-                assert_eq!(data.len(), data_range.len());
+                assert_eq!(data.len(), init.data.len());
                 // If an initializer references an imported memory then
                 // everything will need to be processed in-order anyway to
                 // handle the dynamic limits of the memory specified.
@@ -202,8 +251,8 @@ impl ModuleTranslation<'_> {
                 };
                 let page_size = u64::from(WASM_PAGE_SIZE);
                 let contents = &mut page_contents[memory];
-                let mut page_index = offset / page_size;
-                let mut page_offset = (offset % page_size) as usize;
+                let mut page_index = init.offset / page_size;
+                let mut page_offset = (init.offset % page_size) as usize;
                 let mut data = &data[..];
 
                 while !data.is_empty() {
@@ -248,9 +297,17 @@ impl ModuleTranslation<'_> {
         let mut offset = 0;
         for (memory, pages) in page_contents {
             let mut page_offsets = Vec::with_capacity(pages.len());
-            for (byte_offset, page) in pages {
+            for (page_index, page) in pages {
+                // Skip entirely zero pages since they don't need to participate
+                // in initialization.
+                if page.iter().all(|b| *b == 0) {
+                    continue;
+                }
                 let end = offset + (page.len() as u32);
-                page_offsets.push((byte_offset, offset..end));
+                page_offsets.push(StaticMemoryInitializer {
+                    offset: page_index * u64::from(WASM_PAGE_SIZE),
+                    data: offset..end,
+                });
                 offset = end;
                 self.data.push(page.into());
             }
@@ -259,6 +316,236 @@ impl ModuleTranslation<'_> {
         }
         self.module.memory_initialization = MemoryInitialization::Paged { map };
     }
+
+    /// Similar to the `try_paged_init` method, but attempts to use the
+    /// `MemoryInitialization::Static` variant.
+    ///
+    /// Note that the constraints for `Paged` are the same as those for
+    /// `Static`.
+    ///
+    /// Takes a `page_size` argument in order to ensure that all
+    /// initialization is page-aligned for mmap-ability, and
+    /// `max_image_size_always_allowed` to control how we decide
+    /// whether to use static init.
+    ///
+    /// We will try to avoid generating very sparse images, which are
+    /// possible if e.g. a module has an initializer at offset 0 and a
+    /// very high offset (say, 1 GiB). To avoid this, we use a dual
+    /// condition: we always allow images less than
+    /// `max_image_size_always_allowed`, and the embedder of Wasmtime
+    /// can set this if desired to ensure that static init should
+    /// always be done if the size of the module or its heaps is
+    /// otherwise bounded by the system. We also allow images with
+    /// static init data bigger than that, but only if it is "dense",
+    /// defined as having at least half (50%) of its pages with some
+    /// data.
+    ///
+    /// We could do something slightly better by building a dense part
+    /// and keeping a sparse list of outlier/leftover segments (see
+    /// issue #3820). This would also allow mostly-static init of
+    /// modules that have some dynamically-placed data segments. But,
+    /// for now, this is sufficient to allow a system that "knows what
+    /// it's doing" to always get static init.
+    pub fn try_static_init(&mut self, page_size: u64, max_image_size_always_allowed: u64) {
+        // First try to switch this memory initialization to the `Paged`
+        // variant, if it isn't already. This will perform static bounds checks
+        // and everything and massage it all into a format which is a bit easier
+        // to process here.
+        self.try_paged_init();
+        let map = match &mut self.module.memory_initialization {
+            MemoryInitialization::Paged { map } => map,
+            _ => return,
+        };
+
+        let memory_init_size = |pages: &[StaticMemoryInitializer]| {
+            if pages.len() == 0 {
+                return 0;
+            }
+            let first = &pages[0];
+            let last = &pages[pages.len() - 1];
+            last.offset - first.offset + (last.data.len() as u64)
+        };
+
+        // Perform a check, on all memories, that the memory initialization is
+        // compatible with static memory initialization. The main concern here
+        // is that construction of the memory image shouldn't consume excessive
+        // resources here during compilation. At this point we're already using
+        // paged initialization so we're theoretically using O(data size)
+        // memory already, and we don't want to use excessively more than that
+        // during image construction. Some heuristics are applied here to see if
+        // they're compatible.
+        let mut data = self.data.as_slice();
+        for (_memory_index, pages) in map.iter() {
+            let (memory_data, rest) = data.split_at(pages.len());
+            data = rest;
+
+            // Calculate the total size of data used to initialized this memory
+            // (the sum of all the page sizes), and then also calculate the
+            // actual memory initialization size assuming it's initialized in
+            // one whole chunk in one whole go.
+            let data_size = memory_data.iter().map(|d| d.len()).sum::<usize>() as u64;
+            let memory_init_size = memory_init_size(pages);
+
+            // If the range of memory being initialized is less than twice the
+            // total size of the data itself then it's assumed that static
+            // initialization is ok. This means we'll at most double memory
+            // consumption during the memory image creation process, which is
+            // currently assumed to "probably be ok" but this will likely need
+            // tweaks over time.
+            if memory_init_size < data_size.saturating_mul(2) {
+                continue;
+            }
+
+            // If the memory initialization image is larger than the size of all
+            // data, then we still allow memory initialization if the image will
+            // be of a relatively modest size, such as 1MB here.
+            if memory_init_size < max_image_size_always_allowed {
+                continue;
+            }
+
+            // At this point memory initialization is concluded to be too
+            // expensive to do at compile time so it's entirely deferred to
+            // happen at runtime.
+            return;
+        }
+        assert!(data.is_empty());
+
+        // Here's where we've now committed to changing to static memory. The
+        // memory initialization image is built here from the page data and then
+        // it's converted to a single initializer.
+        let data = mem::replace(&mut self.data, Vec::new());
+        let mut data = data.iter();
+        let mut image_map = PrimaryMap::with_capacity(map.len());
+        let mut offset = 0u32;
+        for (memory_index, pages) in map.iter() {
+            // Allocate the memory image and then fill it in with data. Note
+            // that `pages` should be sorted in increasing order of offsets.
+            let capacity = usize::try_from(memory_init_size(pages)).unwrap();
+            let mut image = Vec::with_capacity(capacity);
+            for page in pages {
+                let image_offset = usize::try_from(page.offset - pages[0].offset).unwrap();
+                assert!(image.len() <= image_offset);
+                image.resize(image_offset, 0u8);
+                image.extend_from_slice(data.next().unwrap());
+            }
+            assert_eq!(image.len(), capacity);
+            assert_eq!(image.capacity(), capacity);
+
+            // Convert the `image` to a single `StaticMemoryInitializer` if it's
+            // not empty.
+            let init = if image.len() > 0 {
+                let data_offset = offset;
+                let len = u32::try_from(image.len()).unwrap();
+                let data_offset_end = data_offset.checked_add(len).unwrap();
+                offset += len;
+
+                // Offset/length should always be page-aligned since our pages
+                // are always wasm pages right now which are 64k and we
+                // otherwise won't work at all on systems larger page sizes.
+                assert!(u64::from(data_offset) % page_size == 0);
+                assert!(u64::from(len) % page_size == 0);
+                self.data.push(image.into());
+
+                Some(StaticMemoryInitializer {
+                    offset: pages[0].offset,
+                    data: data_offset..data_offset_end,
+                })
+            } else {
+                None
+            };
+            let idx = image_map.push(init);
+            assert_eq!(idx, memory_index);
+        }
+        assert!(data.next().is_none());
+        self.data_align = Some(page_size);
+        self.module.memory_initialization = MemoryInitialization::Static { map: image_map };
+    }
+
+    /// Attempts to convert the module's table initializers to
+    /// FuncTable form where possible. This enables lazy table
+    /// initialization later by providing a one-to-one map of initial
+    /// table values, without having to parse all segments.
+    pub fn try_func_table_init(&mut self) {
+        // This should be large enough to support very large Wasm
+        // modules with huge funcref tables, but small enough to avoid
+        // OOMs or DoS on truly sparse tables.
+        const MAX_FUNC_TABLE_SIZE: u32 = 1024 * 1024;
+
+        let segments = match &self.module.table_initialization {
+            TableInitialization::Segments { segments } => segments,
+            TableInitialization::FuncTable { .. } => {
+                // Already done!
+                return;
+            }
+        };
+
+        // Build the table arrays per-table.
+        let mut tables = PrimaryMap::with_capacity(self.module.table_plans.len());
+        // Keep the "leftovers" for eager init.
+        let mut leftovers = vec![];
+
+        for segment in segments {
+            // Skip imported tables: we can't provide a preconstructed
+            // table for them, because their values depend on the
+            // imported table overlaid with whatever segments we have.
+            if self
+                .module
+                .defined_table_index(segment.table_index)
+                .is_none()
+            {
+                leftovers.push(segment.clone());
+                continue;
+            }
+
+            // If this is not a funcref table, then we can't support a
+            // pre-computed table of function indices.
+            if self.module.table_plans[segment.table_index].table.wasm_ty != WasmType::FuncRef {
+                leftovers.push(segment.clone());
+                continue;
+            }
+
+            // If the base of this segment is dynamic, then we can't
+            // include it in the statically-built array of initial
+            // contents.
+            if segment.base.is_some() {
+                leftovers.push(segment.clone());
+                continue;
+            }
+
+            // Get the end of this segment. If out-of-bounds, or too
+            // large for our dense table representation, then skip the
+            // segment.
+            let top = match segment.offset.checked_add(segment.elements.len() as u32) {
+                Some(top) => top,
+                None => {
+                    leftovers.push(segment.clone());
+                    continue;
+                }
+            };
+            let table_size = self.module.table_plans[segment.table_index].table.minimum;
+            if top > table_size || top > MAX_FUNC_TABLE_SIZE {
+                leftovers.push(segment.clone());
+                continue;
+            }
+
+            // We can now incorporate this segment into the initializers array.
+            while tables.len() <= segment.table_index.index() {
+                tables.push(vec![]);
+            }
+            let elements = &mut tables[segment.table_index];
+            if elements.is_empty() {
+                elements.resize(table_size as usize, FuncIndex::reserved_value());
+            }
+
+            let dst = &mut elements[(segment.offset as usize)..(top as usize)];
+            dst.copy_from_slice(&segment.elements[..]);
+        }
+
+        self.module.table_initialization = TableInitialization::FuncTable {
+            tables,
+            segments: leftovers,
+        };
+    }
 }
 
 impl Default for MemoryInitialization {
@@ -312,7 +599,7 @@ impl MemoryInitialization {
     pub fn init_memory(
         &self,
         state: InitMemory<'_>,
-        write: &mut dyn FnMut(MemoryIndex, u64, &Range<u32>) -> bool,
+        write: &mut dyn FnMut(MemoryIndex, &StaticMemoryInitializer) -> bool,
     ) -> bool {
         let initializers = match self {
             // Fall through below to the segmented memory one-by-one
@@ -326,9 +613,23 @@ impl MemoryInitialization {
             // indices are in-bounds.
             MemoryInitialization::Paged { map } => {
                 for (index, pages) in map {
-                    for (page_index, page) in pages {
-                        debug_assert_eq!(page.end - page.start, WASM_PAGE_SIZE);
-                        let result = write(index, *page_index * u64::from(WASM_PAGE_SIZE), page);
+                    for init in pages {
+                        debug_assert_eq!(init.data.end - init.data.start, WASM_PAGE_SIZE);
+                        let result = write(index, init);
+                        if !result {
+                            return result;
+                        }
+                    }
+                }
+                return true;
+            }
+
+            // Like `Paged` above everything's already been validated so this
+            // can simply forward through the data.
+            MemoryInitialization::Static { map } => {
+                for (index, init) in map {
+                    if let Some(init) = init {
+                        let result = write(index, init);
                         if !result {
                             return result;
                         }
@@ -397,7 +698,11 @@ impl MemoryInitialization {
             // The limits of the data segment have been validated at this point
             // so the `write` callback is called with the range of data being
             // written. Any erroneous result is propagated upwards.
-            let result = write(memory_index, start, data);
+            let init = StaticMemoryInitializer {
+                offset: start,
+                data: data.clone(),
+            };
+            let result = write(memory_index, &init);
             if !result {
                 return result;
             }
@@ -460,7 +765,7 @@ impl TablePlan {
     }
 }
 
-/// A WebAssembly table initializer.
+/// A WebAssembly table initializer segment.
 #[derive(Clone, Debug, Serialize, Deserialize)]
 pub struct TableInitializer {
     /// The index of a table to initialize.
@@ -473,6 +778,56 @@ pub struct TableInitializer {
     pub elements: Box<[FuncIndex]>,
 }
 
+/// Table initialization data for all tables in the module.
+#[derive(Clone, Debug, Serialize, Deserialize)]
+pub enum TableInitialization {
+    /// "Segment" mode: table initializer segments, possibly with
+    /// dynamic bases, possibly applying to an imported memory.
+    ///
+    /// Every kind of table initialization is supported by the
+    /// Segments mode.
+    Segments {
+        /// The segment initializers. All apply to the table for which
+        /// this TableInitialization is specified.
+        segments: Vec<TableInitializer>,
+    },
+
+    /// "FuncTable" mode: a single array per table, with a function
+    /// index or null per slot. This is only possible to provide for a
+    /// given table when it is defined by the module itself, and can
+    /// only include data from initializer segments that have
+    /// statically-knowable bases (i.e., not dependent on global
+    /// values).
+    ///
+    /// Any segments that are not compatible with this mode are held
+    /// in the `segments` array of "leftover segments", which are
+    /// still processed eagerly.
+    ///
+    /// This mode facilitates lazy initialization of the tables. It is
+    /// thus "nice to have", but not necessary for correctness.
+    FuncTable {
+        /// For each table, an array of function indices (or
+        /// FuncIndex::reserved_value(), meaning no initialized value,
+        /// hence null by default). Array elements correspond
+        /// one-to-one to table elements; i.e., `elements[i]` is the
+        /// initial value for `table[i]`.
+        tables: PrimaryMap<TableIndex, Vec<FuncIndex>>,
+
+        /// Leftover segments that need to be processed eagerly on
+        /// instantiation. These either apply to an imported table (so
+        /// we can't pre-build a full image of the table from this
+        /// overlay) or have dynamically (at instantiation time)
+        /// determined bases.
+        segments: Vec<TableInitializer>,
+    },
+}
+
+impl Default for TableInitialization {
+    fn default() -> Self {
+        TableInitialization::Segments { segments: vec![] }
+    }
+}
+
 /// Different types that can appear in a module.
 ///
 /// Note that each of these variants are intended to index further into a
@@ -512,8 +867,8 @@ pub struct Module {
     /// The module "start" function, if present.
     pub start_func: Option<FuncIndex>,
 
-    /// WebAssembly table initializers.
-    pub table_initializers: Vec<TableInitializer>,
+    /// WebAssembly table initialization data, per table.
+    pub table_initialization: TableInitialization,
 
     /// WebAssembly linear memory initializer.
     pub memory_initialization: MemoryInitialization,
@@ -527,9 +882,6 @@ pub struct Module {
     /// The map from passive data index (data segment index space) to index in `passive_data`.
     pub passive_data_map: BTreeMap<DataIndex, Range<u32>>,
 
-    /// WebAssembly function names.
-    pub func_names: BTreeMap<FuncIndex, String>,
-
     /// Types declared in the wasm module.
     pub types: PrimaryMap<TypeIndex, ModuleType>,
 
@@ -545,8 +897,15 @@ pub struct Module {
     /// Number of imported or aliased globals in the module.
     pub num_imported_globals: usize,
 
+    /// Number of functions that "escape" from this module may need to have a
+    /// `VMCallerCheckedAnyfunc` constructed for them.
+    ///
+    /// This is also the number of functions in the `functions` array below with
+    /// an `anyfunc` index (and is the maximum anyfunc index).
+    pub num_escaped_funcs: usize,
+
     /// Types of functions, imported and local.
-    pub functions: PrimaryMap<FuncIndex, SignatureIndex>,
+    pub functions: PrimaryMap<FuncIndex, FunctionType>,
 
     /// WebAssembly tables.
     pub table_plans: PrimaryMap<TableIndex, TablePlan>,
@@ -756,11 +1115,30 @@ impl Module {
             EntityIndex::Global(i) => EntityType::Global(self.globals[i]),
             EntityIndex::Table(i) => EntityType::Table(self.table_plans[i].table),
             EntityIndex::Memory(i) => EntityType::Memory(self.memory_plans[i].memory),
-            EntityIndex::Function(i) => EntityType::Function(self.functions[i]),
+            EntityIndex::Function(i) => EntityType::Function(self.functions[i].signature),
             EntityIndex::Instance(i) => EntityType::Instance(self.instances[i]),
             EntityIndex::Module(i) => EntityType::Module(self.modules[i]),
         }
     }
+
+    /// Appends a new function to this module with the given type information,
+    /// used for functions that either don't escape or aren't certain whether
+    /// they escape yet.
+    pub fn push_function(&mut self, signature: SignatureIndex) -> FuncIndex {
+        self.functions.push(FunctionType {
+            signature,
+            anyfunc: AnyfuncIndex::reserved_value(),
+        })
+    }
+
+    /// Appends a new function to this module with the given type information.
+    pub fn push_escaped_function(
+        &mut self,
+        signature: SignatureIndex,
+        anyfunc: AnyfuncIndex,
+    ) -> FuncIndex {
+        self.functions.push(FunctionType { signature, anyfunc })
+    }
 }
 
 /// All types which are recorded for the entirety of a translation.
@@ -792,3 +1170,28 @@ pub struct InstanceSignature {
     /// The name of what's being exported as well as its type signature.
     pub exports: IndexMap<String, EntityType>,
 }
+
+/// Type information about functions in a wasm module.
+#[derive(Debug, Clone, Serialize, Deserialize)]
+pub struct FunctionType {
+    /// The type of this function, indexed into the module-wide type tables for
+    /// a module compilation.
+    pub signature: SignatureIndex,
+    /// The index into the anyfunc table, if present. Note that this is
+    /// `reserved_value()` if the function does not escape from a module.
+    pub anyfunc: AnyfuncIndex,
+}
+
+impl FunctionType {
+    /// Returns whether this function's type is one that "escapes" the current
+    /// module, meaning that the function is exported, used in `ref.func`, used
+    /// in a table, etc.
+    pub fn is_escaping(&self) -> bool {
+        !self.anyfunc.is_reserved_value()
+    }
+}
+
+/// Index into the anyfunc table within a VMContext for a function.
+#[derive(Copy, Clone, PartialEq, Eq, Hash, PartialOrd, Ord, Debug, Serialize, Deserialize)]
+pub struct AnyfuncIndex(u32);
+cranelift_entity::entity_impl!(AnyfuncIndex);
